#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 01/08/2020
           """

import random
import time
from multiprocessing import Process, Queue, current_process, freeze_support


def worker(input, output):
    for func, args in iter(input.get, "STOP"):
        result = calculate(func, args)
        output.put(result)


def calculate(func, args):
    result = func(*args)
    return f"{current_process().name} says that {func.__name__}{args} = {result}"


def mul(a, b):
    time.sleep(0.5 * random.random())
    return a * b


def plus(a, b):
    time.sleep(0.5 * random.random())
    return a + b


def test():
    NUMBER_OF_PROCESSES = 4
    TASKS1 = [(mul, (i, 7)) for i in range(20)]
    TASKS2 = [(plus, (i, 8)) for i in range(10)]

    # Create queues
    task_queue = Queue()
    done_queue = Queue()

    for task in TASKS1:  # Submit tasks
        task_queue.put(task)

    for i in range(NUMBER_OF_PROCESSES):  # Start worker processes
        Process(target=worker, args=(task_queue, done_queue)).start()

    print("Unordered results:")
    for i in range(len(TASKS1)):  # Get and print results
        print("\t", done_queue.get())

    for task in TASKS2:  # Add more tasks using `put()`
        task_queue.put(task)

    for i in range(len(TASKS2)):  # Get and print some more results
        print("\t", done_queue.get())

    for i in range(NUMBER_OF_PROCESSES):  # Tell child processes to stop
        task_queue.put("STOP")


if __name__ == "__main__":
    freeze_support()
    test()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 11-12-2020
           """


async def b():
    import config2

    print(config2.A_CONSTANT)


if __name__ == "__main__":

    async def c():
        await b()

    import asyncio

    asyncio.run(c())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = """
Created on 27/04/2019

@author: cnheider
"""

__all__ = ["SingletonBase", "SingletonMeta", "singleton_2", "singleton"]

import functools
from functools import wraps


class SingletonBase:
    """
    A base class for creating singleton class where all subtypes(Deriavations) should also return the first
    and only
    instantiation of a particular singleton base type, if this property is not wanted consider using the
    SingletonMeta class instead."""

    instance = None

    def __new__(cls, *args, **kwargs):
        if cls.instance is None:
            cls.instance = super().__new__(cls, *args, **kwargs)

        return cls.instance


class SingletonMeta(type):
    """
    Conversely the SingletonBase, this base meta class is used for creating singleton class where all
    subtypes(
    Deriavations) should only
    return
    singleton instantiations of a particular singleton type independantly of subtyping and super-types,
    if this property is not
    wanted
    consider using
    the
    SingletonBase class instead."""

    def __init__(cls, name, bases, namespace):
        super().__init__(name, bases, namespace)
        cls.instance = None

    def __call__(cls, *args, **kwargs):
        if cls.instance is None:
            cls.instance = super().__call__(*args, **kwargs)

        return cls.instance


def singleton(cls):
    """ Use class as singleton. """

    cls.__new_original__ = cls.__new__

    @functools.wraps(cls.__new__)
    def singleton_new(cls, *args, **kw):
        """

        @param cls:
        @type cls:
        @param args:
        @type args:
        @param kw:
        @type kw:
        @return:
        @rtype:"""
        it = cls.__dict__.get("__it__")
        if it is not None:
            return it

        cls.__it__ = it = cls.__new_original__(cls, *args, **kw)
        it.__init_original__(*args, **kw)
        return it

    cls.__new__ = singleton_new
    cls.__init_original__ = cls.__init__
    cls.__init__ = object.__init__

    return cls


def singleton_2(cache_key):
    def inner_fn(fn):
        @wraps(fn)
        def wrapper(self, *args, **kwargs):
            instance = getattr(self, cache_key)
            if instance is not None:
                return instance

            instance = fn(self, *args, **kwargs)
            setattr(self, cache_key, instance)
            return instance

        return wrapper

    return inner_fn


if __name__ == "__main__":

    class SingletonBaseClass(SingletonBase):
        pass

    class S1(SingletonBaseClass):
        pass

    class SingletonBaseMeta(metaclass=SingletonMeta):
        pass

    class S2(SingletonBaseMeta):
        pass

    print(SingletonBaseClass())
    print(SingletonBaseClass())
    print(S1())

    print(SingletonBaseMeta())
    print(SingletonBaseMeta())
    print(S2())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 11-12-2020
           """

from pathlib import Path

from warg.imports import import_warning

import_warning(Path(__file__).with_suffix("").name)

A_CONSTANT = 2
ANOTHER_CONSTANT = 42



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 28-01-2021
           """

from warg import kws_sink, prod, sink


def test_a():
    print((kws_sink("str")))
    print((kws_sink(2)))
    print((kws_sink(2.2)))

    print((prod((2, 2))))

    print((prod((2.2, 2.2))))

    print((prod((2, 2.2))))

    print((prod((2.2, 2))))

    print((sink((2, 2), face=(2.2, 2))))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 02/01/2020
           """

__all__ = [
    "is_positive_and_mod_zero",
    "is_zero_or_mod_zero",
    "is_none_or_zero_or_negative",
    "is_zero_or_mod_below",
    "is_none_or_zero_or_negative_or_mod_zero",
]

from typing import Any

from warg.decorators import drop_unused_kws, passes_kws_to


@drop_unused_kws
def is_positive_and_mod_zero(
    mod: int, counter: int, *, ret: Any = True, alt: Any = False
) -> Any:
    """

    test if mod is positive
    then test if counter % mod is 0
    if both tests are true return ret
    else return alt


    @param mod:
    @param counter:
    @param ret:
    @param alt:
    @return:"""

    return ret if (mod > 0 and (counter % mod == 0)) else alt


@drop_unused_kws
def is_zero_or_mod_below(
    mod: int, below: int, counter: int, *, ret: Any = True, alt: Any = False
) -> Any:
    """

    test if mod is zero or if counter % mod is 0
    if any of the tests are true return ret
    else return alt


    @param below:
    @type below:
    @param mod:
    @param counter:
    @param ret:
    @param alt:
    @return:"""
    return ret if (mod == 0 or (counter % mod < below)) else alt


@drop_unused_kws
def is_zero_or_mod_zero(
    mod: int, counter: int, *, ret: Any = True, alt: Any = False
) -> Any:
    """

    test if mod is zero or if counter % mod is 0
    if any of the tests are true return ret
    else return alt


    @param mod:
    @param counter:
    @param ret:
    @param alt:
    @return:"""
    return ret if (mod == 0 or (counter % mod == 0)) else alt


def is_none_or_zero_or_negative(obj: Any) -> bool:
    """

    @param obj:
    @return:"""
    is_none = obj is None
    is_negative = False
    if isinstance(obj, (int, float)):
        is_negative = obj <= 0

    return is_none or is_negative


@passes_kws_to(is_zero_or_mod_zero)
def is_none_or_zero_or_negative_or_mod_zero(mod: int, counter: int, **kwargs) -> bool:
    """

    @param mod:
    @param counter:
    @param kwargs:
    @return:"""
    return is_none_or_zero_or_negative(mod) or is_zero_or_mod_zero(
        mod, counter, **kwargs
    )


if __name__ == "__main__":
    assert is_zero_or_mod_below(5, 3, 7) == True
    assert is_zero_or_mod_below(5, 2, 4) == False
    for i in range(9):
        print(is_zero_or_mod_zero(1, i))
    for i in range(9):
        print(is_zero_or_mod_zero(2, i))

    print()
    for i in range(8):
        print(is_none_or_zero_or_negative_or_mod_zero(None, i))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import logging
from typing import Any, Mapping, MutableMapping

__author__ = "Christian Heider Nielsen"
__doc__ = ""

__all__ = ["GeneralisedDelayedKwargConstruction", "GDKC"]


class GeneralisedDelayedKwargConstruction(object):
    """
    A generalised class for setting up kwargs for later construction of an instance of an object
    [constructor, args, kwargs]
    """

    def __init__(self, constructor: callable, *args, **kwargs):
        """
        [constructor, args, kwargs]
        :param constructor:
        :param args:
        :param kwargs:"""
        self.constructor: callable = constructor
        assert (
            len(args) < 2
        ), f"Does not support multiple args, only a single mapping type"
        if len(args) == 1:
            assert isinstance(
                args[0], Mapping
            ), f"Arg[0] type is not a mapping type, was {type(args[0])} which is not supported"
        assert not (
            len(kwargs) > 0 and len(args) > 0
        ), f"Does not support both args and kwargs, both supplied args, {args} and {kwargs}"

        if len(args) == 1 and len(kwargs) == 0 and isinstance(args[0], Mapping):
            self.kwargs: MutableMapping = args[0]
        elif len(kwargs) == 1 and next(iter(kwargs.keys())) == "kwargs":
            self.kwargs = kwargs["kwargs"]
        else:
            self.kwargs: MutableMapping = kwargs

    def __enter__(self):
        self.instance = self.__call__()
        return self.instance.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self.instance.__exit__(exc_type, exc_val, exc_tb)

    def __call__(self, *args, **kwargs) -> Any:
        """

        :param args:
        :param kwargs:
        :return:"""

        war = "".join(
            [
                f"Overwriting {k} with the value {v} in construction of {self.constructor}"
                for (k, v), b in zip(kwargs.items(), self.kwargs.keys())
                if k == b
            ]
        )
        if war != "":
            logging.warning(war)

        self.kwargs.update(kwargs)
        try:
            return self.constructor(*args, **self.kwargs)
        except TypeError as e:
            e.args += (f"in construction of {self.constructor}",)
            raise e


GDKC = GeneralisedDelayedKwargConstruction

if __name__ == "__main__":

    class A:
        def __init__(self, *args, **kwargs):
            pass

    def test_not_both():
        GeneralisedDelayedKwargConstruction(A, [1], a=2)

    def test_kw():
        GeneralisedDelayedKwargConstruction(A, a=2)

    def test_mapping():
        GeneralisedDelayedKwargConstruction(A, {"a": 2})

    def test_mapping_and_args_fail():
        GeneralisedDelayedKwargConstruction(A, {"a": 2}, 1)

    def test_mapping_and_args_fail_inv():
        GeneralisedDelayedKwargConstruction(A, 1, {"a": 2})

    test_kw()
    test_mapping()
    try:
        test_mapping_and_args_fail()
        raise Exception
    except:
        pass
    try:
        test_mapping_and_args_fail_inv()
        raise Exception
    except:
        pass
    test_not_both()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import numpy

__author__ = "Christian Heider Nielsen"
__doc__ = ""

__all__ = [
    "IterDictItemsMixin",
    "IterDictKeysMixin",
    "IterDictValuesMixin",
    "OrdinalIndexingDictMixin",
]


class IterDictItemsMixin:
    """
    Mixin class for iterating kw pairs in a class instance __dict__"""

    def __iter__(self):
        for attr, value in self.__dict__.items():
            yield attr, value


class IterDictKeysMixin:
    """
    Mixin class for iterating only the keys of a class instance __dict__"""

    def __iter__(self):
        for attr in self.__dict__.keys():
            yield attr


class IterDictValuesMixin:
    """
    Mixin class for iterating only the values of a class instance __dict__"""

    def __iter__(self):
        for value in self.__dict__.values():
            yield value


class OrdinalIndexingDictMixin:
    """
    Mixin class for indexing a class instance __dict__ (SortedDict) with both integer (ordinal) indexing or
    key:str attributes (non-ordinal) access."""

    def __getitem__(self, item):
        if isinstance(item, int):
            a = self.__dict__
            b = numpy.array(list(a.values()))
            return b[item]
        else:
            return self.__dict__[item]


if __name__ == "__main__":

    def asd():
        class IDTM(OrdinalIndexingDictMixin):
            pass

        a = IDTM()
        a.a = 2
        a.b = 3
        assert a[0] == 2
        assert a[1] == 3

    def asdij():
        class IASD(IterDictValuesMixin):
            pass

        a = IASD()
        a.b = 1
        a.c = 2
        a.d = 3

        for ca in a:
            print(ca)

    asd()
    asdij()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 01/08/2020
           """

__all__ = []


def test_import():
    import warg

    print(warg.__version__)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from typing import (
    Any,
    ItemsView,
    Iterable,
    KeysView,
    List,
    MutableMapping,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    ValuesView,
)

import sorcery
from sorcery.core import node_name

__author__ = "Christian Heider Nielsen"

__all__ = ["NamedOrderedDictionary", "NOD"]

LOCALS = (
    "as_list",
    "as_dict",
    "as_tuples",
    "as_flat_tuples",
    "add_unnamed_arg",
    "dict_of",
    "keys",
    "update",
    "__setattr__",
)


class IllegalAttributeKey(Exception):
    """
    An exception for when a deemed illegal attribute key was being overwritten"""

    def __init__(self, key, type_: Type):
        Exception.__init__(
            self,
            f'Overwriting of attribute "{key}" on type "{type_.__name__}" is not allowed',
        )


T = TypeVar("T", bound="NamedOrderedDictionary")


class NamedOrderedDictionary(MutableMapping):
    """

    Usage:

    .. code-block:: python

    nodict = NamedOrderedDictionary()
    nodict.paramA = 'str_parameter'
    nodict.paramB = 10
    assert nodict.paramA == 'str_parameter'
    assert nodict.paramB == 10


    .. code-block:: python

    nodict = NamedOrderedDictionary()
    nodict['paramA'] = 10
    assert nodict.paramA == 10

    .. code-block:: python

    nodict = NamedOrderedDictionary({'paramA': 'str_parameter', 'paramB': 10})
    assert nodict.paramA == 'str_parameter'
    assert nodict.paramB == 10

    .. code-block:: python

    nodict = NamedOrderedDictionary(paramA='str_parameter', paramB=10)
    assert nodict.paramA == 'str_parameter'
    assert nodict.paramB == 10

    .. code-block:: python

    nodict = NamedOrderedDictionary('str_parameter', 10)
    assert nodict.arg0 == 'str_parameter'
    assert nodict.arg1 == 10

    .. code-block:: python

    arg0,arg1 = NamedOrderedDictionary('str_parameter', 10).as_list()
    assert arg0 == 'str_parameter'
    assert arg1 == 10

    As with dictionaries you can use the `update()` method.

    .. code-block:: python
    nodict = NamedOrderedDictionary()
    nodict.update({'paramA': 20, 'paramB': 'other_param', 'paramC': 5.0})
    assert nodict.paramA == 20
    assert nodict.paramB == 'other_param'

    .. code-block:: python

    nodict = NamedOrderedDictionary('str_parameter', 10)
    nodict.update({'arg1': 20, 'arg0': 'other_param'})
    assert nodict.arg0 == 'other_param'
    assert nodict.arg1 == 20

    .. code-block:: python

    nodict = NamedOrderedDictionary(paramA='str_parameter', paramB=10)
    nodict.update(20,'other_param')
    assert nodict.paramB == 'other_param'
    assert nodict.paramA == 20
    """

    # __slots__ = ('_unnamed_arg_i','__dict__')

    # _unnamed_arg_i = 0

    def __init__(self, *args, **kwargs) -> None:
        # super().__init__(**kwargs)
        if len(args) == 1 and isinstance(args[0], dict):
            args_dict = args[0]
        else:
            args_dict = {}
            if len(args) == 1 and isinstance(args[0], Iterable):
                args = args[0]
            for arg in args:
                args_dict[id(arg)] = arg

        args_dict.update(kwargs)
        self.update(args_dict or {})

    def as_list(self) -> list:
        """

        :return:
        :rtype:"""
        return list(self.__dict__.values())

    def as_dict(self) -> dict:
        """

        :return:
        :rtype:"""
        return self.__dict__

    def as_tuples(self) -> List[Tuple[Any, Any]]:
        """

        :return:
        :rtype:"""
        return [(k, v) for (k, v) in self.__dict__.items()]

    def as_flat_tuples(self) -> List[Tuple]:
        """

        :return:
        :rtype:"""
        return [(k, *v) for (k, v) in self.__dict__.items()]

    def add_unnamed_arg(self, arg: Any) -> None:
        """

        :param arg:
        :type arg:"""
        self.__dict__[f"arg{id(arg)}"] = arg

    @staticmethod
    @sorcery.spell  # TODO: MAY BE BROKEN!
    def nod_of(frame_info, *args, **kwargs) -> T:
        """Instead of:

        {'foo': foo, 'bar': bar, 'spam': thing()}

        or:

        dict(foo=foo, bar=bar, spam=thing())

        write:

        NOD.dict_of(foo, bar, spam=thing())

        In other words, returns a NamedOrderedDictionary with an item for each argument,
        where positional arguments use their names as keys,
        and keyword arguments do the same as in the usual dict constructor.

        The positional arguments can be any of:

        - plain variables,
        - attributes, or
        - subscripts (square bracket access) with string literal keys

        So the following:

        NOD.dict_of(spam, x.foo, y['bar'])

        is equivalent to:

        NOD.dict_of(spam=spam, foo=x.foo, bar=y['bar'])

        *args are not allowed.

        :rtype: object
        """
        nod = NamedOrderedDictionary()

        for arg, value in zip(frame_info.call.args[-len(args) :], args):
            try:
                arg_key = node_name(arg)
                nod[arg_key] = value
            except TypeError:
                nod.add_unnamed_arg(value)

        nod.update(kwargs)

        return nod

    def __getattr__(self, item: Any) -> Any:
        if item == "__deepcopy__":
            return super().__getattribute__(item)
        return self.__dict__[item]

    def __len__(self) -> int:
        return len(self.__dict__)

    def __setattr__(self, key: Any, value: Any) -> None:
        if key in LOCALS:
            raise IllegalAttributeKey(key, type_=NamedOrderedDictionary)

        if key == "__dict__":
            super().__setattr__(key, value)
        else:
            self.__dict__[key] = value

    def __getitem__(self, key: Any) -> Any:
        """
        NOTE getting a tuple is a unique key

        :param key:
        :type key:
        :return:
        :rtype:"""
        if isinstance(key, slice):
            keys = list(self.__dict__.keys())[key]
            return [self.__dict__[a] for a in keys]
        elif isinstance(key, KeysView):
            # assert set(self.__dict__.keys()).issuperset(key)
            return [self.__dict__[a] for a in key]
        return self.__dict__[key]

    def __setitem__(self, key: Any, value: Any) -> None:
        """
        NOTE setting a tuple is a unique key

        :param key:
        :type key:
        :param value:
        :type value:"""
        if isinstance(key, slice):
            keys = list(self.__dict__.keys())[key]
            if isinstance(value, Sequence):
                assert len(keys) == len(
                    value
                ), f"number of keys {len(keys)} are not equal values {len(value)}"
                for a, v in zip(keys, value):
                    self.__dict__[a] = v
            else:
                for a in keys:
                    self.__dict__[a] = value
        elif isinstance(key, KeysView):
            # assert set(self.__dict__.keys()).issuperset(key)
            # assert isinstance(value,Sequence), f'values must be of type Sequence, was {type(value)},' \
            #                                f' distribution is not supported'
            if isinstance(value, Sequence):
                assert len(key) == len(
                    value
                ), f"number of keys {len(key)} are not equal values {len(value)}"
                for a, v in zip(key, value):
                    self.__dict__[a] = v
            else:
                for a in key:
                    self.__dict__[a] = value
        else:
            self.__dict__[key] = value

    def __delitem__(self, key) -> None:
        del self.__dict__[key]

    def keys(self) -> KeysView:
        """

        :return:
        :rtype:"""
        return self.__dict__.keys()

    def items(self) -> ItemsView:
        """

        :return:
        :rtype:"""
        return self.__dict__.items()

    def values(self) -> ValuesView:
        """

        :return:
        :rtype:"""
        return self.__dict__.values()

    def __contains__(self, item) -> bool:
        return item in self.__dict__

    def __iter__(self):
        for key, value in self.__dict__.items():
            yield key, value

    def __repr__(self):
        items = self.items()
        print_str = f"{self.__class__.__name__}("
        if len(items) > 0:
            for key, value in items:
                print_str += f"'{key}': {value}, "
            print_str = print_str[:-2]
        print_str += ")"
        return print_str

    def update(self, *args: Any, **kwargs: Any) -> T:
        """
        Merge two attributes, overriding any repeated keys from
        the `items` parameter.

        :returns self

        Args:
        items (dict): Python dictionary containing updated values."""

        if len(args) == 1 and isinstance(args[0], dict):
            args_dict = args[0]
        else:
            args_dict = {}
            a = list(self.__dict__.keys())
            for arg, key in zip(args, a):
                args_dict[key] = arg

        args_dict.update(kwargs)

        for key in args_dict:
            if key in LOCALS:
                raise IllegalAttributeKey(key, type_=NamedOrderedDictionary)

        self.__dict__.update(args_dict)
        return self

    def __add__(self, other) -> T:
        cop = self.__dict__.copy()
        if isinstance(other, NamedOrderedDictionary):
            for k in other.keys():
                if k in cop:
                    cop[k] += other.__dict__[k]
                else:
                    cop[k] = other.__dict__[k]
        elif isinstance(other, Iterable):
            for arg in other:
                self.add_unnamed_arg(arg)
        else:
            self.add_unnamed_arg(other)
        return NOD(cop)

    def __sub__(self, other) -> T:
        cop = self.__dict__.copy()
        if isinstance(other, NamedOrderedDictionary):

            for k in other.keys():
                if k in cop:
                    cop[k] -= other.__dict__[k]
            return NOD(cop)
        else:
            raise ArithmeticError(f"Can not subtract {type(other)} from {type(self)}")

    def __truediv__(self, other) -> Any:
        if isinstance(other, (str, int)):
            return self.get(other)
        else:
            raise ArithmeticError(f"Can not access with {type(other)} in {type(self)}")

    def __matmul__(self, other):
        if isinstance(other, (str, int)):
            return self.get(other)
        else:
            raise ArithmeticError(f"Can not access with {type(other)} in {type(self)}")

    def __floordiv__(self, other) -> Any:
        return self.__truediv__(other)

    def __setstate__(self, state) -> None:
        self.__dict__ = state

    def __getstate__(self) -> dict:
        return self.__dict__

    def __call__(self, *args) -> List:
        return [self.get(a) for a in args]


NOD = NamedOrderedDictionary

if __name__ == "__main__":

    nodict = NamedOrderedDictionary()
    nodict.paramA = "str_parameter"
    nodict.paramB = 10
    assert nodict.paramA == "str_parameter"
    assert nodict.paramB == 10
    from copy import deepcopy, copy

    b = copy(nodict)
    print(b)
    assert b.paramB == 10
    assert b.paramB == nodict.paramB

    a = deepcopy(nodict)
    print(a)
    assert a.paramB == 10
    assert a.paramB == nodict.paramB



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 15-12-2020
           """

from multiprocessing import Pool


def f(x):
    import config1

    print(config1.A_CONSTANT)
    return x * x


if __name__ == "__main__":
    with Pool(5) as p:
        print(p.map(f, [1, 2, 3]))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 11/12/2019
           """

__all__ = ["PostInit"]


class PostInit(type):
    """
    define a new metaclass which overrides the "__call__" function"""

    def __call__(cls, *args, **kwargs):
        """
        Called when you call a class type constructor()"""
        obj = type.__call__(cls, *args, **kwargs)
        if hasattr(obj, "__post_init__"):
            obj.__post_init__(*args, **kwargs)
        return obj


if __name__ == "__main__":

    class SAD(metaclass=PostInit):
        def __init__(self):
            print("init")

        def __post_init__(self):
            print("post_init")

    SAD()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import pytest

from warg.data_structures.named_ordered_dictionary import IllegalAttributeKey, NOD

__author__ = "Christian Heider Nielsen"


def test_attribute_assignment():
    nodict = NOD()
    nodict.paramA = "str_parameter"
    nodict.paramB = 10
    assert nodict.paramA == "str_parameter"
    assert nodict.paramB == 10


def test_dict_arg_construction():
    nodict = NOD({"paramA": "str_parameter", "paramB": 10})
    assert nodict.paramA == "str_parameter"
    assert nodict.paramB == 10


def test_update_and_set_attr_update():
    nodict = NOD()
    nodict.update({"paramA": 20, "paramB": "other_param", "paramC": 5.0})
    nodict.paramA = 10
    assert nodict.paramA == 10
    assert nodict.paramB == "other_param"


def test_set_item_update():
    nodict = NOD()
    nodict["paramA"] = 10
    assert nodict.paramA == 10


def test_as_list_args_and_kwargs():
    vals = [1, 3, 5]
    nodict = NOD(10, val2=2, *vals)
    nolist = nodict.as_list()
    assert nolist[0] == 10
    assert nodict.val2 == 2
    assert nolist[-2] == vals[-1]


def test_kv_unmap():
    nodict = NOD("str_parameter", 10)
    odict = {**nodict}
    assert nodict.as_list()[0] == list(odict.values())[0]


def test_update_attr_with_kwargs():
    nodict = NOD("str_parameter", 10)
    nodict.update(arg1=20, arg0="other_param")
    assert nodict.arg0 == "other_param"
    assert nodict.arg1 == 20


def test_update_attr_with_dict_arg():
    nodict = NOD("str_parameter", 10)
    nodict.update({"arg1": 20, "arg0": "other_param"})
    assert nodict.arg0 == "other_param"
    assert nodict.arg1 == 20


def test_no_existing_attr():
    nodict = NOD(paramA="str_parameter", paramB=10)
    nodict.update(20, "other_param")
    assert nodict.paramB == "other_param"
    assert nodict.paramA == 20
    assert nodict.get("paramC") is None


def test_existing_map_unpacking():
    nodict = NOD(paramA="str_parameter", paramB=10)
    nodict.update(20, "other_param")
    nodict = NOD(paramC="str_parameter", **nodict)
    assert nodict.paramB == "other_param"
    assert nodict.paramA == 20
    assert nodict.get("paramC") is not None


def test_as_list():
    arg0, arg1 = NOD("str_parameter", 10).as_list()
    assert arg0 == "str_parameter"
    assert arg1 == 10


def test_access_operators_NOD():
    arg0 = "str_parameter"
    arg1 = 10
    l = NOD(arg0, arg1)
    assert l[id(arg1)] == arg1
    assert l.__getattr__(id(arg1)) == arg1
    assert l[id(arg0)] == arg0
    assert l / id(arg0) == arg0
    assert l @ id(arg0) == arg0


def test_new_attr_on_existing():
    a = NOD(4, 2)
    a.list = "a"
    assert a.list == "a"


def test_addition_arithmetic():
    a = NOD(4, 2)
    b = a + a
    assert b.as_list() == [8, 4]


def test_illegal_overwrite():
    with pytest.raises(IllegalAttributeKey) as exc_info:
        NOD(update=2)

    assert exc_info.type is IllegalAttributeKey


def test_slice_get():
    a = NOD(4, 2)
    b = a + a
    assert b[:1] == [8]
    assert b[-1:] == [4]


def test_slice_get_adv():
    a = NOD(4, 2)
    b = a[:1] + a[0:]
    assert b[:1] == [4]
    assert b[-1:] == [2]


def test_slice_set():
    a = NOD(4, 2)
    a[:1] = [8]
    assert a.as_list() == [8, 2]


def test_slice_all():
    a = NOD(4, 2)
    b = a + a + a
    b[:1] = [8]
    assert b.as_list() == [8, 6]


def test_sorcery():
    from sorcery import dict_of

    def ret1():
        return 1

    arg1 = ret1()
    columns = dict_of(arg1)
    assert columns["arg1"] == arg1


def test_access_operators_NOD_of():
    list_rep = NOD("str_parameter", 10).as_list()
    arg0 = list_rep[0]
    arg1 = list_rep[1]
    columns = NOD.nod_of(arg1, dsa=arg0)
    assert columns["arg1"] == arg1
    assert columns.arg1 == arg1
    assert columns["dsa"] == arg0
    assert columns / "dsa" == arg0
    assert columns @ "dsa" == arg0
    assert id(columns / "dsa") == id(columns["dsa"])


def test_multiple_return_assignment_to_contstruction():
    arg0, arg1 = ("sta", 1)
    columns = NOD.nod_of(arg1, arg0)
    assert columns.arg1 == 1
    assert columns.arg0 == "sta"


def test_access_operators_single_explicit_naming():
    arg0 = "str_parameter"
    columns = NOD.nod_of(dsa=arg0)
    assert columns["dsa"] == arg0
    assert columns / "dsa" == arg0
    assert id(columns / "dsa") == id(columns["dsa"])


def test_access_operators_single_no_naming():
    dsa = "str_parameter"
    columns = NOD.nod_of(dsa)
    assert columns["dsa"] == dsa
    assert columns / "dsa" == dsa
    assert id(columns / "dsa") == id(columns["dsa"])


def test_access_operators_no_multi_return_no_variable_name_direct_inference():
    arg0 = "str_parameter"
    columns = NOD.nod_of("sas", dsa=arg0)
    print(columns)
    assert columns[:1] == ["sas"]
    assert columns.as_list()[0] == "sas"
    assert columns.as_list()[1] == arg0
    assert columns["dsa"] == arg0
    assert columns / "dsa" == arg0
    assert id(columns / "dsa") == id(columns["dsa"])


def test_nested():
    cfg = NOD()

    cfg.MODEL = NOD()

    cfg.MODEL.DEVICE = "cuda"
    # match default boxes to any ground truth with jaccard overlap higher than a threshold (0.5)
    cfg.MODEL.THRESHOLD = 0.5
    cfg.MODEL.NUM_CLASSES = 21
    # Hard negative mining
    cfg.MODEL.NEG_POS_RATIO = 3
    cfg.MODEL.CENTER_VARIANCE = 0.1
    cfg.MODEL.SIZE_VARIANCE = 0.2

    cfg.ADSA = 203182

    print(cfg)


def test_copy():
    from copy import deepcopy, copy

    nodict = NOD()
    nodict.paramA = "str_parameter"
    nodict.paramB = 10

    b = copy(nodict)
    print(b)
    assert b.paramB == 10
    assert b.paramB == nodict.paramB

    a = deepcopy(nodict)
    print(a)
    assert a.paramB == 10
    assert a.paramB == nodict.paramB


def test_distribute_keys_view():
    nodict = NOD()
    nodict.paramA = 2
    nodict.paramB = 10

    nodict.paramC = 5
    nodict[nodict.keys()] = 4

    assert nodict.paramA == 4
    assert nodict.paramB == 4
    assert nodict.paramC == 4


def test_distribute_slice():
    nodict = NOD()
    nodict.paramA = 2
    nodict.paramB = 10
    nodict.paramC = 5
    nodict.paramD = 5

    nodict[:3] = 4

    assert nodict.paramA == 4
    assert nodict.paramB == 4
    assert nodict.paramC == 4
    assert nodict.paramD == 5


def test_multi_key_index():
    nodict = NOD()
    nodict.paramA = 2
    nodict.paramB = 10
    nodict.paramC = 5
    nodict.paramD = 5

    nodict["paramA", "paramB"] = 4

    assert nodict.paramA == 2
    assert nodict.paramB == 10
    assert nodict.paramC == 5
    assert nodict.paramD == 5

    assert nodict["paramA", "paramB"] == 4


def test_call_index():
    nodict = NOD()
    nodict.paramA = 2
    nodict.paramB = 10
    nodict.paramC = 5
    nodict.paramD = 5
    nodict[6] = 1
    nodict[1] = 1
    nodict[2] = 11

    assert nodict.paramA == 2
    assert nodict.paramB == 10
    assert nodict.paramC == 5
    assert nodict.paramD == 5

    assert nodict("paramA", "paramB", 6) == [2, 10, 1]
    assert nodict("paramC") == [5]
    assert nodict(6, "paramD", 1) == [1, 5, 1]
    assert nodict(2) == [11]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""
          Automatically generates a look up data

           Created on 06/03/2020
           """

from time import sleep, time
from typing import Iterable, Mapping, Set, Tuple

from warg.decorators.hashing import make_hash

global_table = {}

__all__ = ["add_lut", "look_up", "look_up_args", "look_up_kws"]


def add_lut(f: callable):
    """

    :param f:
    :type f:
    :return:
    :rtype:"""
    global_table[f] = {}
    return f


def look_up(f, *args, **kwargs):
    """

    :param f:
    :type f:
    :param args:
    :type args:
    :param kwargs:
    :type kwargs:
    :return:
    :rtype:"""
    ag_hash = hash(args) + make_hash(kwargs)
    if f in global_table:
        if ag_hash in global_table[f]:
            return global_table[f][ag_hash]

        res = global_table[f][ag_hash] = f(*args, **kwargs)
        return res

    global_table[f] = {}
    res = global_table[f][ag_hash] = f(*args, **kwargs)
    return res


def look_up_args(f, *args):
    """

    :param f:
    :type f:
    :param args:
    :type args:
    :return:
    :rtype:"""
    if f in global_table:
        if args in global_table[f]:
            return global_table[f][args]

        res = global_table[f][args] = f(*args)
        return res

    global_table[f] = {}
    res = global_table[f][args] = f(*args)
    return res


def look_up_kws(f, **kws):
    """

    :param f:
    :type f:
    :param kws:
    :type kws:
    :return:
    :rtype:"""
    kw_hash = make_hash(kws)
    if f in global_table:
        if kw_hash in global_table[f]:

            return global_table[f][kw_hash]

        res = global_table[f][kw_hash] = f(**kws)
        return res

    global_table[f] = {}
    res = global_table[f][kw_hash] = f(**kws)
    return res


def precompute_lut(
    f: callable, arg_sets: Set[Tuple[Iterable, Mapping]], *, verbose=False
) -> callable:
    """

    :param f:
    :type f:
    :param arg_sets:
    :type arg_sets:
    :param verbose:
    :type verbose:
    :return:
    :rtype:"""
    for arg_set, kws_set in arg_sets:
        res = look_up(f, *arg_set, **kws_set)
        if verbose:
            print(f"precompute {f},{arg_set},{kws_set}->{res}")
    return f


def precompute_lut_args(
    f: callable, arg_sets: Set[Iterable], *, verbose=False
) -> callable:
    """

    :param f:
    :type f:
    :param arg_sets:
    :type arg_sets:
    :param verbose:
    :type verbose:
    :return:
    :rtype:"""
    for arg_set in arg_sets:
        res = look_up_args(f, *arg_set)
        if verbose:
            print(f"precompute {f},{arg_sets}->{res}")
    return f


def precompute_lut_kws(
    f: callable, arg_sets: Set[Mapping], *, verbose=False
) -> callable:
    """

    :param f:
    :type f:
    :param arg_sets:
    :type arg_sets:
    :param verbose:
    :type verbose:
    :return:
    :rtype:"""
    for arg_set in arg_sets:
        res = look_up_kws(f, **arg_set)
        if verbose:
            print(f"precompute {f},{arg_sets}->{res}")
    return f


def precompute_lut_dec(arg_sets, *, verbose=False):
    """

    :param arg_sets:
    :type arg_sets:
    :param verbose:
    :type verbose:
    :return:
    :rtype:"""

    def rdec(f):
        """

        :param f:
        :type f:
        :return:
        :rtype:"""
        precompute_lut(f, arg_sets, verbose=verbose)
        return f

    return rdec


def precompute_lut_args_dec(arg_sets, *, verbose: bool = False):
    """

    :param arg_sets:
    :type arg_sets:
    :param verbose:
    :type verbose:
    :return:
    :rtype:"""

    def rdec(f):
        """

        :param f:
        :type f:
        :return:
        :rtype:"""
        precompute_lut_args(f, arg_sets, verbose=verbose)
        return f

    return rdec


if __name__ == "__main__":

    @add_lut
    def my_function(foo, bar):
        """

        :param foo:
        :type foo:
        :param bar:
        :type bar:
        :return:
        :rtype:"""
        return foo + bar

    result = look_up_args(my_function, 0, 8)
    print(result)
    result = look_up_args(my_function, 8, 0)
    print(result)

    @precompute_lut_args_dec(list(zip(range(9), reversed(range(9)))))
    def my_function2(foo, bar):
        """

        :param foo:
        :type foo:
        :param bar:
        :type bar:
        :return:
        :rtype:"""
        sleep(0.1)
        return bar * foo

    a = time()
    look_up_args(my_function2, 0, 8)
    print(time() - a)
    a = time()
    look_up_args(my_function2, 8, 0)
    print(time() - a)

    result = look_up(my_function2, 14, bar=21)
    print(result)

    result = look_up(my_function2, 14, bar=21)
    print(result)

    result = look_up(my_function2, 104, bar=2)
    print(result)

    result = look_up_args(my_function, 1, 2)
    print(result)

    result = look_up_args(my_function, 1, 2)
    print(result)

    result = look_up_args(my_function, 12, 2)
    print(result)

    result = look_up_args(my_function2, 1, 2)
    print(result)

    result = look_up_kws(my_function2, foo=1, bar=2)
    print(result)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 15-12-2020
           """

if __name__ == "__main__":

    def f():
        import config1

        print(config1.A_CONSTANT)
        import config1

        print(config1.ANOTHER_CONSTANT)

    f()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from warg.metas.singleton import SingletonBase, SingletonMeta

__author__ = "Christian Heider Nielsen"
__doc__ = r"""
           """


def test_singleton_class():
    class SingletonBaseClass(SingletonBase):
        pass

    class S1(SingletonBaseClass):
        pass

    a = SingletonBaseClass()
    b = S1()

    assert id(a) == id(b) == id(SingletonBaseClass())


def test_singleton_meta():
    class SingletonBaseMeta(metaclass=SingletonMeta):
        pass

    class S2(SingletonBaseMeta):
        pass

    a = SingletonBaseMeta()
    b = S2()

    assert (id(a) == id(SingletonBaseMeta())) != (id(b) == S2())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 14/01/2020
           """

__all__ = [
    "kws_sink",
    "sink",
    "prod",
    "collate_first_dim",
    "call_identity",
    "args_sink",
    "identity",
]

import operator
from functools import reduce
from typing import Any, Dict, Iterable, Tuple, Union

from warg import drop_unused_kws


def identity(a: Any) -> Any:
    return a


@drop_unused_kws
def kws_sink(*args) -> Tuple[Any, ...]:
    """
    Returns args without any modification what so ever. Drops kws
    :param x:
    :return:"""
    return args


def call_identity(*args, **kwargs) -> Tuple[Tuple[Any], Dict[str, Any]]:
    """

    :param args:
    :param kwargs:
    :return:"""
    return args, kwargs


def args_sink(*args, **kwargs) -> Dict[str, Any]:
    """

    :param args:
    :param kwargs:
    :return:"""
    return kwargs


def sink(*args, **kwargs) -> None:
    """
    Returns None, but accepts everything

    :param args:
    :param kwargs:
    :return:"""
    return None


def prod(iterable: Iterable[Union[int, float]]) -> Union[int, float]:
    """
    Calculate the product of the a Iterable of int or floats
    :param iterable:
    :return:"""
    return reduce(operator.mul, iterable, 1)


def collate_first_dim(batch: Iterable) -> tuple:
    """

    :param batch:
    :return:"""
    return tuple(zip(*batch))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pytest

from warg.pooled_queue_processor import PooledQueueProcessor, PooledQueueTask
from warg.functions import identity

__author__ = "Christian Heider Nielsen"


class Square(PooledQueueTask):
    def call(self, i, *args, **kwargs):
        return i * 2


class Exc(PooledQueueTask):
    def call(self, *args, **kwargs):
        raise NotImplementedError


@pytest.mark.skip

def test_integration_success():
    task = Square()

    with PooledQueueProcessor(
        task, [2], fill_at_construction=True, max_queue_size=10
    ) as processor:
        for a, _ in zip(processor, range(30)):
            pass
            # print(a)


@pytest.mark.skip

def test_integration_func():
    task = identity

    with PooledQueueProcessor(task, [2], max_queue_size=10) as processor:
        for a, _ in zip(processor, range(30)):
            pass
            # print(a)


@pytest.mark.skip
def test_lambda_func():
    task = lambda x: x

    with PooledQueueProcessor(task, [2], max_queue_size=10) as processor:
        for a, _ in zip(processor, range(30)):
            pass
            # print(a)


@pytest.mark.skip
def test_integration_except():
    task = Exc()

    with pytest.raises(NotImplementedError) as exc_info:
        task()  # TODO: MP does not work in pytest
        processor = PooledQueueProcessor(task, [2], max_queue_size=10, blocking=True)
        for a, _ in zip(processor, range(30)):
            pass
            # print(a)

    assert exc_info.type is NotImplementedError


@pytest.mark.skip
#
def test_integration_except_ctx():
    task = Exc()

    with pytest.raises(NotImplementedError) as exc_info:
        task()  # TODO: MP does not work in pytest
        with PooledQueueProcessor(task, [2], max_queue_size=10) as processor:
            for a, _ in zip(processor, range(30)):
                pass
                # print(a)

    assert exc_info.type is NotImplementedError


if __name__ == "__main__":
    test_lambda_func()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import pytest

from warg.gdkc import GeneralisedDelayedKwargConstruction

__author__ = "Christian Heider Nielsen"
__doc__ = r"""
            Much like the partial wrapper from functools, GeneralisedDelayedKwargConstruction provides a
            way of finishing constrution of class instance or call of function but lets you change the
            kwargs until actual call.
           """


class A:
    def __init__(self, *args, **kwargs):
        self.args_a = args
        self.kwargs_a = kwargs

    def something(self):
        print(self.kwargs_a)

    def something_else(self, *args):
        print(args, self.kwargs_a)

    def another(self, *args, **kwargs):
        print(args, self.kwargs_a, kwargs)

    def clearly_something(self, *args, **kwargs):
        print(self.args_a, args, self.kwargs_a, kwargs)


def test_not_both():
    try:
        GeneralisedDelayedKwargConstruction(A, [1], a=2)
    except Exception as e:
        assert isinstance(e, AssertionError)


def test_kw():
    GeneralisedDelayedKwargConstruction(A, a=2)


def test_mapping():
    da = GeneralisedDelayedKwargConstruction(A, {"a": 2})
    a = da(99)
    a.something()
    a.clearly_something(231, b="52")


def test_overwriting_warning(caplog):
    ssc = GeneralisedDelayedKwargConstruction(A, {"a": 2})

    ssc(a=1)

    assert "Overwriting" in caplog.text


def test_mapping_and_args_fail():
    with pytest.raises(AssertionError) as exc_info:
        GeneralisedDelayedKwargConstruction(A, {"a": 2}, 1)

    assert exc_info.type is AssertionError


def test_mapping_and_args_fail_inv():
    with pytest.raises(AssertionError) as exc_info:
        GeneralisedDelayedKwargConstruction(A, 1, {"a": 2})

    assert exc_info.type is AssertionError



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 18/07/2020
           """

import numbers
from typing import Sequence, Tuple, Union

__all__ = [
    "Number",
    "Reals",
    "Numbers",
    "StrictNumbers",
    "Single",
    "Double",
    "Triple",
    "Quad",
    "Quint",
]

Number = Union[int, float]
Reals = Sequence[Union[numbers.Real, "Reals"]]
Numbers = Sequence[Union[Number, "Numbers"]]
StrictNumbers = Union[
    Union[Sequence[int], "StrictNumbers"], Union[Sequence[float], "StrictNumbers"]
]

Single = Tuple[Number]
Double = Tuple[Number, Number]
Triple = Tuple[Number, Number, Number]
Quad = Tuple[Number, Number, Number, Number]
Quint = Tuple[Number, Number, Number, Number, Number]

if __name__ == "__main__":

    def stest():
        assert (
            isinstance(1, Number.__args__)
            and isinstance(1.1, Number.__args__)
            and not isinstance(complex(1, 1), Number.__args__)
        )
        # assert isinstance(list(range(2)), Numbers.__args__) and isinstance((float(i) for i in range(2)), Numbers.__args__) and isinstance((1,2.0), Numbers.__args__)
        # assert isinstance(list(range(2)), StrictNumbers.__args__) and isinstance((float(i) for i in range(2)), StrictNumbers.__args__) and not isinstance((1,2.0), StrictNumbers.__args__)
        assert (
            isinstance(1, numbers.Real)
            and isinstance(1.1, numbers.Real)
            and not isinstance(complex(2, 2), numbers.Real)
        )
        # assert isinstance(list(range(2)), Reals.__args__) and isinstance((float(i) for i in range(2)), Reals.__args__) and isinstance((1,2.0), Reals.__args__)

    stest()
